<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:r="http://www.istandaarden.nl/validatie/1_0/rapport/schema/1_0">
	<xsl:output method="xml" version="1.0" encoding="UTF-8" indent="yes" />

	<!-- Emit the (absolute or relative to the given context) XPath expression for this node; uses indexing when needed context is not checked currently -->
	<xsl:template name="printXPath">
		<xsl:param name="pElement" select="." />
		<!-- Call recursive unless root; we have to use this way of detecting root for Java -->
		<xsl:if test="$pElement/ancestor::*"><!--
			--><xsl:call-template name="printXPath">
				<xsl:with-param name="pElement" select="$pElement/.." />
			</xsl:call-template><!-- 
			-->/<!--
		--></xsl:if><!-- Print self -->
		<xsl:value-of select="local-name($pElement)" /><!-- Print child selector if applicable -->
		<xsl:if test="($pElement/preceding-sibling::*|$pElement/following-sibling::*)[local-name()=local-name($pElement)]">
			<xsl:value-of select="concat('[',count($pElement/preceding-sibling::*[local-name()=local-name($pElement)])+1,']')" />
		</xsl:if>
	</xsl:template>

	<!-- Evaluate an XPath expression relative to the given context; note: only supports node selectors and indexing, and the expression must start 
		with a / -->
	<xsl:template name="evaluateXPath">
		<xsl:param name="pPath" select="." />
		<xsl:param name="pContext" select="/" />

		<xsl:choose>
			<xsl:when test="string-length($pPath) > 0">
				<!-- Strip leading / -->
				<xsl:variable name="vPath" select="substring($pPath,2)" />
				<!-- Extract first node selector -->
				<xsl:variable name="vNodeTest">
					<xsl:choose>
						<xsl:when test="not(contains($vPath, '/'))">
							<xsl:value-of select="$vPath" />
						</xsl:when>
						<xsl:otherwise>
							<xsl:value-of select="substring-before($vPath, '/')" />
						</xsl:otherwise>
					</xsl:choose>
				</xsl:variable>
				<!-- Extract node position -->
				<xsl:variable name="vPosition">
					<xsl:choose>
						<xsl:when test="not(contains($vNodeTest, '['))">
							1
						</xsl:when>
						<xsl:otherwise>
							<xsl:value-of select="substring-after(substring-before($vNodeTest, ']'), '[')" />
						</xsl:otherwise>
					</xsl:choose>
				</xsl:variable>
				<!-- Extract just the node name without position -->
				<xsl:variable name="vNameTest">
					<xsl:choose>
						<xsl:when test="not(contains($vNodeTest, '['))">
							<xsl:value-of select="$vNodeTest" />
						</xsl:when>
						<xsl:otherwise>
							<xsl:value-of select="substring-before($vNodeTest, '[')" />
						</xsl:otherwise>
					</xsl:choose>
				</xsl:variable>
				<!-- Call recursively -->
				<xsl:call-template name="evaluateXPath">
					<xsl:with-param name="pPath" select="substring-after($pPath, $vNodeTest)" />
					<xsl:with-param name="pContext"
						select="$pContext/*[local-name() = $vNameTest and count(preceding-sibling::*[local-name() = $vNameTest]) = $vPosition - 1]" />
				</xsl:call-template>
			</xsl:when>
			<xsl:otherwise>
				<!-- Just print the node contents if we are done -->
				<xsl:value-of select="normalize-space($pContext/text())" />
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>

<!-- Add an error element for violating pRule at pElements -->
	<xsl:template name="addError">
		<xsl:param name="pRule">
			FOUT
		</xsl:param>
		<xsl:param name="pElements" select="." />
		<r:Fout>
			<r:Code>
				<xsl:value-of select="normalize-space($pRule)" />
			</r:Code>
			<xsl:for-each select="$pElements">
				<r:Locatie><!--
					-->/<!--
					--><xsl:call-template name="printXPath">
						<xsl:with-param name="pElement" select="."></xsl:with-param>
					</xsl:call-template>
				</r:Locatie>
			</xsl:for-each>
		</r:Fout>
	</xsl:template>

<!-- Add a error of $pRule for $pElement if $pElement doesn't satisfy the 11-proef -->
	<xsl:template name="check11proef">
		<xsl:param name="pElement" select="." />
		<xsl:param name="pRule">
			FOUT
		</xsl:param>
		<xsl:variable name="digit1" select="floor(number($pElement/text()) div 1) mod 10" />
		<xsl:variable name="digit2" select="floor(number($pElement/text()) div 10) mod 10" />
		<xsl:variable name="digit3" select="floor(number($pElement/text()) div 100) mod 10" />
		<xsl:variable name="digit4" select="floor(number($pElement/text()) div 1000) mod 10" />
		<xsl:variable name="digit5" select="floor(number($pElement/text()) div 10000) mod 10" />
		<xsl:variable name="digit6" select="floor(number($pElement/text()) div 100000) mod 10" />
		<xsl:variable name="digit7" select="floor(number($pElement/text()) div 1000000) mod 10" />
		<xsl:variable name="digit8" select="floor(number($pElement/text()) div 10000000) mod 10" />
		<xsl:variable name="digit9" select="floor(number($pElement/text()) div 100000000) mod 10" />

		<xsl:if
			test="((9 * $digit9 + 8 * $digit8 + 7 * $digit7 + 6 * $digit6 + 5 * $digit5 + 4 * $digit4 + 3 * $digit3 + 2 * $digit2 - 1 * $digit1) mod 11) != 0">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pElement" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>

<!-- Helper for check testvalue; checks lists of values in XSLT1 -->
	<xsl:template name="split">
		<xsl:param name="pElement" select="." />
		<xsl:param name="pValue" />

		<xsl:choose>
			<xsl:when test="string-length(normalize-space($pValue)) &gt; 0">
				<xsl:variable name="vNextItem" select="normalize-space(substring-before(concat($pValue, ','), ','))" />
				<xsl:choose>
					<xsl:when test="string-length(normalize-space($vNextItem)) > 1 and
					                            substring(normalize-space($vNextItem), 1, 1) = '!' and
					                            normalize-space(substring(normalize-space($vNextItem), 2)) != normalize-space($pElement/text())">
						<xsl:value-of select="true()" />
					</xsl:when>
					<xsl:when test="normalize-space($pElement/text()) = $vNextItem">
						<xsl:value-of select="true()" />
					</xsl:when>
					<xsl:otherwise>
						<xsl:call-template name="split">
							<xsl:with-param name="pElement" select="$pElement" />
							<xsl:with-param name="pValue" select="substring-after($pValue, ',')" />
						</xsl:call-template>
					</xsl:otherwise>
				</xsl:choose>
			</xsl:when>
			<xsl:otherwise>
				<xsl:value-of select="false()" />
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>
	
	<xsl:template name="callrecursive">
		<xsl:param name="pT" />
		<xsl:param name="pValues" />
		<xsl:param name="pText" />
		<xsl:param name="pPattern" />
		
		<xsl:choose>
			<xsl:when test="string-length($pT) = 1 and string-length(translate($pT, $pValues, '')) = 0">
				<xsl:call-template name="patternmatch">
					<xsl:with-param name="pText" select="$pText" />
					<xsl:with-param name="pPattern" select="$pPattern" />
				</xsl:call-template>
			</xsl:when>
			<xsl:otherwise>
				<xsl:value-of select="false()" />
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>
	
	<xsl:template name="patternmatch">
		<xsl:param name="pText" />
		<xsl:param name="pPattern" />
		<xsl:choose>
			<xsl:when test="(string-length($pText) = 0 and string-length($pPattern) != 0) or (string-length($pText) != 0 and string-length($pPattern) = 0)">
				<xsl:value-of select="false()" />
			</xsl:when>
			<xsl:when test="string-length($pText) = 0"> <!-- Nothing left to match -->
				<xsl:value-of select="true()" />
			</xsl:when>
			<xsl:otherwise>
				<xsl:variable name="t" select="substring($pText, 1, 1)" />
				<xsl:variable name="p" select="substring($pPattern, 1, 1)" />
				<xsl:variable name="vText" select="substring($pText, 2)" />
				<xsl:variable name="vPattern" select="substring($pPattern, 2)" />
				<xsl:choose>
					<xsl:when test="$p = '.'">
						<xsl:call-template name="patternmatch">
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$p = '\' and string-length($vPattern) != 0">
						<xsl:call-template name="callrecursive">
							<xsl:with-param name="pT" select="$t" />
							<xsl:with-param name="pValues" select="substring($vPattern, 1, 1)" />
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="substring($vPattern, 2)" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$p = '1'">
						<xsl:call-template name="callrecursive">
							<xsl:with-param name="pT" select="$t" />
							<xsl:with-param name="pValues">123456789</xsl:with-param>
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$p = '0'">
						<xsl:call-template name="callrecursive">
							<xsl:with-param name="pT" select="$t" />
							<xsl:with-param name="pValues">0123456789</xsl:with-param>
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$p = 'a'">
						<xsl:call-template name="callrecursive">
							<xsl:with-param name="pT" select="$t" />
							<xsl:with-param name="pValues">abcdefghijklmnopqrstuvwxyz</xsl:with-param>
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$p = 'A'">
						<xsl:call-template name="callrecursive">
							<xsl:with-param name="pT" select="$t" />
							<xsl:with-param name="pValues">ABCDEFGHIJKLMNOPQRSTUVWXYZ</xsl:with-param>
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$p = 'z' or $p = 'Z'">
						<xsl:call-template name="callrecursive">
							<xsl:with-param name="pT" select="$t" />
							<xsl:with-param name="pValues">abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ</xsl:with-param>
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:when test="$t = $p">
						<xsl:call-template name="patternmatch">
							<xsl:with-param name="pText" select="$vText" />
							<xsl:with-param name="pPattern" select="$vPattern" />
						</xsl:call-template>
					</xsl:when>
					<xsl:otherwise>
						<xsl:value-of select="false()" />
					</xsl:otherwise>
				</xsl:choose>
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>
	
	<!-- Very simple pattern matching; check that pElement matches pPattern; '0' = any number, '1' = 1-9, 'a' = lowercase letter, 'A' = uppercase letter, 'z'/'Z' = letter, '.' any character; unmatched characters are matched literally, characters can be escaped using \ to be matched literally -->
	<xsl:template name="testpattern">
		<xsl:param name="pElement" select="." />
		<xsl:param name="pPattern" select="''" />
		<xsl:variable name="vText" select="$pElement/text()" />
		<xsl:call-template name="patternmatch">
			<xsl:with-param name="pText" select="$vText" />
			<xsl:with-param name="pPattern" select="$pPattern" />
		</xsl:call-template>
	</xsl:template>

	<!-- Test that $pElement matches $pValues.  $pValues can be '' (non-existant or empty), '*' (existant and non-empty), '#' (existant node), '!#' (non-existant node), '#<pattern>' (matching <pattern> according to testpattern), or a (list of) allowed/disallowed values (disallow by prefixing by !) -->
	<xsl:template name="testvalue">
		<xsl:param name="pElement" select="." />
		<xsl:param name="pValues" select="''" />
		<xsl:choose>
			<xsl:when test="normalize-space($pValues) = ''">
				<xsl:choose>
					<xsl:when test="not($pElement/.)">
						<xsl:value-of select="true()" />
					</xsl:when>
					<xsl:when test="normalize-space($pElement/text()) = ''">
						<xsl:value-of select="true()" />
					</xsl:when>
					<xsl:otherwise>
						<xsl:value-of select="false()" />
					</xsl:otherwise>
				</xsl:choose>
			</xsl:when>
			<xsl:when test="normalize-space($pValues) = '*'">
				<xsl:choose>
					<xsl:when test="normalize-space($pElement/text()) != ''">
						<xsl:value-of select="true()" />
					</xsl:when>
					<xsl:otherwise>
						<xsl:value-of select="false()" />
					</xsl:otherwise>
				</xsl:choose>
			</xsl:when>
			<xsl:when test="normalize-space($pValues) = '#'">
				<xsl:choose>
					<xsl:when test="$pElement/.">
						<xsl:value-of select="true()" />
					</xsl:when>
					<xsl:otherwise>
						<xsl:value-of select="false()" />
					</xsl:otherwise>
				</xsl:choose>
			</xsl:when>
			<xsl:when test="normalize-space($pValues) = '!#'">
				<xsl:choose>
					<xsl:when test="$pElement/.">
						<xsl:value-of select="false()" />
					</xsl:when>
					<xsl:otherwise>
						<xsl:value-of select="true()" />
					</xsl:otherwise>
				</xsl:choose>
			</xsl:when>
			<xsl:when test="string-length(normalize-space($pValues)) > 1 and substring(normalize-space($pValues), 1, 1) = '#'">
				<xsl:call-template name="testpattern">
					<xsl:with-param name="pElement" select="$pElement" />
					<xsl:with-param name="pPattern" select="substring(normalize-space($pValues), 2)" />
				</xsl:call-template>
			</xsl:when>
			<xsl:otherwise>
				<xsl:call-template name="split">
					<xsl:with-param name="pElement" select="$pElement" />
					<xsl:with-param name="pValue" select="$pValues" />
				</xsl:call-template>
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>


	<!-- Check that $pThis is empty iff $pOther is and add an error for $pRule otherwise-->
	<xsl:template name="checkempty">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pOther" select="." />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:variable name="thisEmpty">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pThis" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:variable name="otherEmpty">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pOther" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:if test="$thisEmpty = 'false' and $otherEmpty = 'true'">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis|$pOther" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	
	<!-- Check pThis comes (non-strictly) after pOther; must be dates in the forst YYYY-MM-DD -->
	<xsl:template name="checkafter">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pOther" select="." />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:if test="not($pThis) or not($pOther) or translate($pOther/text(), '-', '') &gt; translate($pThis/text(), '-', '')">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis|$pOther" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	
	<!-- Check pThis comes (strictly) after pOther; must be dates in the forst YYYY-MM-DD -->
	<xsl:template name="checkafterstrict">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pOther" select="." />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:if test="not($pThis) or not($pOther) or translate($pOther/text(), '-', '') &gt;= translate($pThis/text(), '-', '')">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis|$pOther" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	
	<!-- Check that $pThis matches $pValues (same as testvalue) and add a violation of $pRule otherwise -->
	<xsl:template name="checkvalue">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pValues" select="''" />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:variable name="thisCorrect">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pThis" />
				<xsl:with-param name="pValues" select="$pValues" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:if test="$thisCorrect = 'false'">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	
	<!-- Check that $pThis matches $pValues (same as testvalue) and add a violation of $pRule if it does -->
	<xsl:template name="checknvalue">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pValues" select="''" />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:variable name="thisCorrect">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pThis" />
				<xsl:with-param name="pValues" select="$pValues" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:if test="$thisCorrect = 'true'">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>

	<!-- Check that $pThis matches $pValue1 (same as testvalue) and $pOther matches $pValue2; add a violation of $pRule if not (pThis => pOther), i.e., if this matches and other doesn't -->
	<xsl:template name="checkimplication">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pValue1" select="''" />
		<xsl:param name="pOther" select="." />
		<xsl:param name="pValue2" select="''" />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:variable name="thisCorrect">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pThis" />
				<xsl:with-param name="pValues" select="$pValue1" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:variable name="otherCorrect">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pOther" />
				<xsl:with-param name="pValues" select="$pValue2" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:if test="$thisCorrect = 'true' and $otherCorrect = 'false'">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis|$pOther" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>

	<!-- Check that $pThis matches $pValue1 (same as testvalue) and $pOther matches $pValue2; add a violation of $pRule if they don't both match or none matches -->
	<xsl:template name="checksame">
		<xsl:param name="pThis" select="." />
		<xsl:param name="pValue1" select="''" />
		<xsl:param name="pOther" select="." />
		<xsl:param name="pValue2" select="''" />
		<xsl:param name="pRule">
			FAIL
		</xsl:param>
		
		<xsl:variable name="thisCorrect">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pThis" />
				<xsl:with-param name="pValues" select="$pValue1" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:variable name="otherCorrect">
			<xsl:call-template name="testvalue">
				<xsl:with-param name="pElement" select="$pOther" />
				<xsl:with-param name="pValues" select="$pValue2" />
			</xsl:call-template>
		</xsl:variable>

		<xsl:if test="$thisCorrect != $otherCorrect">
			<xsl:call-template name="addError">
				<xsl:with-param name="pRule" select="$pRule" />
				<xsl:with-param name="pElements" select="$pThis|$pOther" />
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	
	<xsl:template match="*[local-name()='BerichtCode']" mode="header">
		<r:BerichtCode>
			<xsl:value-of select="text()" />
		</r:BerichtCode>
	</xsl:template>
	<xsl:template match="*[local-name()='BerichtVersie']" mode="header">
		<r:BerichtVersie>
			<xsl:value-of select="text()" />
		</r:BerichtVersie>
	</xsl:template>
	<xsl:template match="*[local-name()='BerichtSubversie']" mode="header">
		<r:BerichtSubversie>
			<xsl:value-of select="text()" />
		</r:BerichtSubversie>
	</xsl:template>
	<xsl:template match="*[local-name()='Identificatie']" mode="header">
		<r:Identificatie>
			<xsl:value-of select="text()" />
		</r:Identificatie>
	</xsl:template>
	<xsl:template match="*[local-name()='Dagtekening']" mode="header">
		<r:Dagtekening>
			<xsl:value-of select="text()" />
		</r:Dagtekening>
	</xsl:template>
	<xsl:template match="*[local-name()='BerichtIdentificatie']" mode="header">
		<r:BerichtIdentificatie>
			<xsl:apply-templates select="*" mode="header" />
		</r:BerichtIdentificatie>
	</xsl:template>

	<xsl:template match="*|@*|text()" mode="check" />
	<xsl:template match="*|@*|text()" mode="header">
		<xsl:apply-templates select="*" mode="header" />
	</xsl:template>
	<xsl:template match="*|@*|text()" mode="traverse">
		<xsl:apply-templates select="*" mode="traverse" />
		<xsl:apply-templates select="*" mode="check" />
	</xsl:template>
	<xsl:template match="*[local-name() = 'Bericht']">
		<r:Rapport>
			<r:Header>
				<xsl:apply-templates select="//*[local-name() = 'Header']" mode="header" />
			</r:Header>
			<r:Fouten>
				<xsl:apply-templates select="*" mode="traverse" />
			</r:Fouten>
		</r:Rapport>
	</xsl:template>
</xsl:stylesheet>
